# encoding: utf-8

from .utils import find_version

# Coindesk API client settings
API_CLIENT_VERSION = find_version('__init__.py')
API_PROTOCOL = 'https'
API_HOST = 'api.coindesk.com'
API_PATH = '/v1/bpi'
API_CURRENTPRICE_DATA_TYPE = 'currentprice'
API_HISTORICAL_DATA_TYPE = 'historical'
API_SUPPORTED_CURRENCIES_DATA_TYPE = 'supported-currencies'
API_CURRENTPRICE_ENDPOINT = f'{API_CURRENTPRICE_DATA_TYPE}' + '{currency}.json'
API_HISTORICAL_ENDPOINT = f'{API_HISTORICAL_DATA_TYPE}/close.json'
API_SUPPORTED_CURRENCIES_ENDPOINT = f'{API_SUPPORTED_CURRENCIES_DATA_TYPE}.json'
API_ENDPOINTS = {
    API_CURRENTPRICE_DATA_TYPE: API_CURRENTPRICE_ENDPOINT,
    API_HISTORICAL_DATA_TYPE: API_HISTORICAL_ENDPOINT,
    API_SUPPORTED_CURRENCIES_DATA_TYPE: API_SUPPORTED_CURRENCIES_ENDPOINT,
}

# Coindesk API valid parameters
INDEX_PARAM = 'index'
CURRENCY_PARAM = 'currency'
START_PARAM = 'start'
END_PARAM = 'end'
FOR_PARAM = 'for'
VALID_DATA_TYPES = [
    API_CURRENTPRICE_DATA_TYPE,
    API_HISTORICAL_DATA_TYPE,
]
VALID_CURRENTPRICE_PARAMS = [
    CURRENCY_PARAM,
]
VALID_HISTORICAL_PARAMS = [
    INDEX_PARAM,
    CURRENCY_PARAM,
    START_PARAM,
    END_PARAM,
    FOR_PARAM,
]
VALID_INDEX = [
    'USD',
    'CNY',
]
VALID_FOR = [
    'yesterday',
]

# Coindesk API client request configuration parameters
REQUEST_MAX_RETRIES = 10
REQUEST_MAX_TIMEOUT = 30
REQUEST_HEADERS = {
    'Accept': 'application/json',
    'Accept-Language': 'en-US',
    'Cache-Control': 'no-cache',
    'Connection': 'close',
    'X-API-client-version': API_CLIENT_VERSION
}



# encoding: utf-8

import asyncio
import json
import math
import re
from collections import OrderedDict
from json import JSONDecodeError
from logging import getLogger
from logging.config import fileConfig
from os.path import dirname, join

import aiohttp
import jsonschema
import requests
from aiohttp import ClientResponse, ClientSession
from furl import furl as URL
from jsonschema import SchemaError, ValidationError
from requests.exceptions import RequestException

from . import settings, utils
from .decorators import async_event_loop
from .exceptions import (CoindeskAPIClientError,
                         CoindeskAPIHttpRequestError,
                         CoindeskAPIHttpResponseError)

# Custom logger for client module
fileConfig(join(dirname(dirname(__file__)), 'logging.cfg'))
logger = getLogger(__name__)


class CoindeskAPIHttpRequest(object):
    """
    Enable Coindesk API http request.
    """

    def __init__(self, retries: int = 10, redirects: bool = True, timeout: int = 5, backoff: bool = True):
        """
        Initialize Coindesk API http request making.

        :param int retries: number of request attempts before failing.
        :param bool redirects: enable/disable http verbs redirection.
        :param int timeout: seconds before request timeout.
        :param bool backoff: enable/disable http request retry backoff.
        """
        self._retries = retries
        self._redirects = redirects
        self._timeout = timeout
        self._backoff = backoff

    def __str__(self):
        """
        Represent class via params string.

        :return str: class representation.
        """
        classname = self.__class__.__name__
        return f'<{classname} - Coindesk api request>'

    @classmethod
    def start(cls, retries: int = 10, redirects: bool = True, timeout: int = 5, backoff: bool = True):
        """
        Get Coindesk API http request instance.

        :param int retries: number of request attempts before failing.
        :param bool redirects: enable/disable http verbs redirection.
        :param int timeout: seconds before request timeout.
        :param bool backoff: enable/disable http request retry backoff.
        :return cls: CoindeskAPICient class instance.
        """
        retries, redirects, timeout, backoff = cls.validate(retries, redirects, timeout, backoff)
        return cls(retries, redirects, timeout, backoff)

    @staticmethod
    def validate(retries: int, redirects: bool, timeout: int, backoff: bool):
        """
        Validate initialization parameters.

        :param int retries: number of request attempts before failing.
        :param bool redirects: enable/disable http verbs redirection.
        :param int timeout: seconds before request timeout.
        :param bool backoff: enable/disable http request retry backoff.
        :return tuple: validated params.
        """
        retries = utils.validate_retries(retries)
        redirects = utils.validate_redirects(redirects)
        timeout = utils.validate_timeout(timeout)
        backoff = utils.validate_backoff(backoff)
        return retries, redirects, timeout, backoff

    @property
    def retries(self):
        """
        Get number of request attempts.

        :return int: number of request attempts before failing.
        """
        return self._retries

    @retries.setter
    def retries(self, retries: int):
        """
        Set number of request attempts.

        :param int retries: number of request attempts before failing.
        """
        retries = utils.validate_retries(retries)
        self._retries = retries

    @property
    def redirects(self):
        """
        Check if requests redirects are allowed.

        :return bool: http verbs redirection status.
        """
        return self._redirects

    @redirects.setter
    def redirects(self, redirects: bool):
        """
        Set/unset redirects.

        :param bool redirects: enable/disable http verbs redirection.
        """
        redirects = utils.validate_redirects(redirects)
        self._redirects = redirects

    @property
    def timeout(self):
        """
        Get number of seconds for request timeout.
        """
        return self._timeout

    @timeout.setter
    def timeout(self, timeout: int):
        """
        Set number of seconds for request timeout.

        :param int timeout: seconds before request timeout.
        """
        timeout = utils.validate_timeout(timeout)
        self._timeout = timeout

    @property
    def backoff(self):
        """
        Check if requests retries backoff is enabled.

        :return bool: request retries backoff status.
        """
        return self._backoff

    @backoff.setter
    def backoff(self, backoff: bool):
        """
        Set/unset request backoff.

        :param bool backoff: enable/disable http requests backoff.
        """
        backoff = utils.validate_backoff(backoff)
        self._backoff = backoff

    @async_event_loop
    async def get(self, url: str, raw: bool = False):
        """
        Retrieve response object/data from Coindesk API url.

        :param str url: api resource locator.
        :param bool raw: enable/disable api response parsing.
        :return *: api http raw response or response data.
        """
        options = self._get_request_options()
        async with aiohttp.ClientSession() as session:
            response = await self._http_request(session, url, options)
        self._check_response_status(response)
        return response if raw else await self._get_json_response(response)

    def _get_request_options(self):
        """
        Return http get request option parameters.
        """
        return {
            'headers': OrderedDict(settings.REQUEST_HEADERS),
            'allow_redirects': self.redirects,
            'timeout': self.timeout
        }

    async def _http_request(self, session: ClientSession, url: str, options: dict):
        """
        Make asynchronous http request to Coindesk API.

        :param obj session: client session.
        :param str url: optional query parameters.
        :param dict options: http request options.
        :return obj: http response object.
        """
        for retry in range(1, self.retries + 1):
            try:
                return await session.get(url, **options)
            except RequestException as err:
                timeout = self._wait_exp_backoff(retry) if self.backoff else 0
                logger.error(f'[CoindeskAPIHttpRequest] Retry {retry} request. {err.args[0]}.')
                logger.error(f'[CoindeskAPIHttpRequest] Waiting {timeout} ms.')
                await asyncio.sleep(timeout / 10)
        else:
            msg = f'No response from Coindesk API url {url}.'
            logger.error(f'[CoindeskAPIHttpRequest] Request error. {msg}')
            raise CoindeskAPIHttpRequestError(msg)

    def _wait_exp_backoff(self, retry: int):
        """
        Calculate exponential backoff time.

        :param int retries: number of request attempts before failing.
        :return int: time to wait between http requests retries.
        """
        return math.pow(2, retry)

    def _check_response_status(self, response: ClientResponse):
        """
        Verify http response status code.

        :param int status: http response status code.
        """
        status = str(response.status)
        msg = f'Status code {status} - {response.reason}.'
        if status.startswith('1'):
            logger.info(f'[CoindeskAPIHttpRequest] Request info. {msg}')
        elif status.startswith('2'):
            logger.info(f'[CoindeskAPIHttpRequest] Request success. {msg}')
        elif status.startswith('3'):
            logger.info(f'[CoindeskAPIHttpRequest] Request redirect. {msg}')
        elif status.startswith('4'):
            logger.error(f'[CoindeskAPIHttpRequest] Client error. {msg}')
            raise CoindeskAPIHttpRequestError(msg)
        elif status.startswith('5'):
            logger.error(f'[CoindeskAPIHttpRequest] Server error. {msg}')
            raise CoindeskAPIHttpRequestError(msg)

    async def _get_json_response(self, response: ClientResponse):
        """
        Return response json data format.

        :param obj response: http response object.
        :return json: response json data.
        """
        try:
            data = await response.json(content_type=None)
        except JSONDecodeError as err:
            msg = f'Could not decode json data. {err.args[0]}.'
            logger.error(f'[CoindeskAPIHttpRequest] Request error. {msg}')
            raise CoindeskAPIHttpRequestError(msg)
        logger.info('[CoindeskAPIHttpRequest] JSON data decoded.')
        return data


class CoindeskAPIClient(CoindeskAPIHttpRequest):
    """
    Enable Coindesk API use.
    """

    def __init__(self, data_type: str = None, params: dict = None, retries: int = 10,
                 redirects: bool = True, timeout: int = 5, backoff: bool = True):
        """
        Initialize Coindesk API client.

        :param str data_type: type of data to fetch (currentprice, historical).
        :param dict params: optional url query parameters.
        :param int retries: number of request attempts before failing.
        :param bool redirects: enable/disable http verbs redirection.
        :param int timeout: seconds before request timeout.
        :param bool backoff: enable/disable http request retry backoff.
        """
        if params is None: params = {}
        super(CoindeskAPIClient, self).__init__(retries, redirects, timeout, backoff)
        self._data_type = data_type
        self._api_endpoint = self._construct_api_endpoint(data_type, params)

    def __str__(self):
        """
        Represent class via params string.

        :return str: class representation.
        """
        classname = self.__class__.__name__
        endpoint = self._api_endpoint
        return f'<{classname} - Coindesk api client\nendpoint: {endpoint}>'

    @classmethod
    def start(cls, data_type: str = None, params: dict = None, retries: int = 10,
              redirects: bool = True, timeout: int = 5, backoff: bool = True):
        """
        Get Coindesk API client instance.

        :param str data_type: type of data to fetch (currentprice, historical).
        :param dict params: optional url query parameters.
        :param int retries: number of request attempts before failing.
        :param bool redirects: enable/disable http verbs redirection.
        :param int timeout: seconds before request timeout.
        :param bool backoff: enable/disable http request retry backoff.
        :return cls: CoindeskAPICient class instance.
        """
        if params is None: params = {}
        data_type = utils.validate_data_type(data_type)
        params = utils.validate_params(data_type, params)
        retries, redirects, timeout, backoff = cls.validate(retries, redirects, timeout, backoff)
        return cls(data_type, params, retries, redirects, timeout, backoff)

    def _construct_api_endpoint(self, data_type: str, params: dict):
        """
        Get Coindesk api endpoint.

        :param str data_type: type of data to fetch (currentprice, historical).
        :param dict params: optional query parameters.
        :return str: Coindesk api endpoint for correspondig data resource.
        """
        scheme, host, path = self._get_api_url_components()
        resource = settings.API_ENDPOINTS.get(data_type)
        if data_type == settings.API_CURRENTPRICE_DATA_TYPE:
            currency_param = params.pop(settings.CURRENCY_PARAM, '')
            currency = f'/{currency_param}' if currency_param else currency_param
            resource = resource.format(currency=currency)
        path = f'{path}/{self._clean_api_component(resource)}'
        api_endpoint = URL(scheme=scheme, host=host, path=path, args=params)
        utils.validate_url(api_endpoint.url)
        return api_endpoint

    def _get_api_url_components(self):
        """
        Get Coindesk api base url components.

        :return list: Coindesk api base url components.
        """
        components = settings.API_PROTOCOL, settings.API_HOST, settings.API_PATH
        return [self._clean_api_component(component) for component in components]

    def _clean_api_component(self, component: str):
        """
        Clean Coindesk api base url components.

        :param str component: Coindesk api base url component.
        :return str: Coindesk api base url cleaned component.
        """
        return re.sub('://', '', component).strip('/')

    @property
    def data_type(self):
        """
        Get Coindesk data type resource.
        """
        return self._data_type

    @data_type.setter
    def data_type(self, data_type: str):
        """
        Set Coindesk API client data type and corresponding api path attribute.

        :param str data_type: type of data to fetch (currentprice, historical).
        """
        self._data_type = utils.validate_data_type(data_type)
        self._api_endpoint = self._construct_api_endpoint(data_type, {})

    @property
    def url(self):
        """
        Get Coindesk api endpoint.
        """
        return self._api_endpoint.url

    @property
    def origin(self):
        """
        Get Coindesk api endpoint origin.
        """
        return self._api_endpoint.origin

    @property
    def path(self):
        """
        Get Coindesk api endpoint path.
        """
        return str(self._api_endpoint.path)

    @property
    def params(self):
        """
        Get Coindesk api endpoint optional query parameters.
        """
        if self.data_type == settings.API_CURRENTPRICE_DATA_TYPE:
            if self.path.endswith('currentprice.json'): return []
            currency = self._api_endpoint.path.segments[-1].rstrip('.json')
            return [('currency', currency)]
        elif self.data_type == settings.API_HISTORICAL_DATA_TYPE:
            return self._api_endpoint.query.params.items()

    @params.setter
    def params(self, params: dict):
        """
        Set Coindesk API client optinal query parameters.
        This will overwrite all existing params.

        :param dict params: optional url query parameters.
        """
        params = utils.validate_params(self.data_type, params)
        if self.data_type == settings.API_CURRENTPRICE_DATA_TYPE:
            self._api_endpoint = self._construct_api_endpoint(self.data_type, params)
        elif self.data_type == settings.API_HISTORICAL_DATA_TYPE:
            self._api_endpoint.query.params = params

    def has_param(self, key: str):
        """
        Check if Coindesk API endpoint has certain param.

        :param str key: query param name.
        :return bool: true/false has param response.
        """
        has_param = key in self._api_endpoint.query.params
        if self.data_type == settings.API_CURRENTPRICE_DATA_TYPE and key == settings.CURRENCY_PARAM:
            has_param = not self.path.endswith('currentprice.json')
        return has_param

    def add_param(self, param: dict):
        """
        Add Coindesk API client optinal query parameter.

        :param dict param: optional url query parameter.
        """
        param = utils.validate_params(self.data_type, param)
        if self.data_type == settings.API_CURRENTPRICE_DATA_TYPE:
            self._api_endpoint = self._construct_api_endpoint(self.data_type, param)
        elif self.data_type == settings.API_HISTORICAL_DATA_TYPE:
            key = list(param.keys())[0]
            if self.has_param(key): self.delete_param(key)
            self._api_endpoint.add(args=param)

    def add_many_params(self, params: dict):
        """
        Add many Coindesk API client optinal query parameters.

        :param dict params: optional url query parameters.
        """
        for key, value in params.items():
            self.add_param({key: value})

    def delete_param(self, key: str):
        """
        Delete Coindesk API client optinal query parameter.

        :param str param: optional url query parameter.
        """
        deleted_param = None
        if self.data_type == settings.API_CURRENTPRICE_DATA_TYPE:
            if not self.path.endswith('currentprice.json'):
                segments = self._api_endpoint.path.segments
                self._api_endpoint.path = f'{"/".join(segments[:-1])}.json'
                deleted_param = segments[-1].rstrip('.json')
        elif self.data_type == settings.API_HISTORICAL_DATA_TYPE:
            deleted_param = self._api_endpoint.query.params.pop(param, None)

        if deleted_param is None:
            msg = f'Query parameter {param} does not exist.'
            logger.warning(f'[CoindeskAPIClient] Delete param. {msg}')
        return deleted_param

    def delete_many_params(self, keys: list):
        """
        Delete many Coindesk API client optinal query parameters.

        :param list params: optional url query parameters.
        """
        for key in keys:
            self.delete_param(key)

    @property
    def valid_params(self):
        """
        Get Coindesk valid query parameters for api endpoint.
        """
        if self.data_type == settings.API_CURRENTPRICE_DATA_TYPE:
            return settings.VALID_CURRENTPRICE_PARAMS
        elif self.data_type == settings.API_HISTORICAL_DATA_TYPE:
            return settings.VALID_HISTORICAL_PARAMS
        else:
            msg = f'Uncorrect data type setup for {self.data_type}.'
            logger.warning(f'[CoindeskAPICient] Data type error. {msg}')
            return None

    def get_supported_currencies(self):
        """
        Get Coindesk valid currencies list.
        """
        scheme, host, path = self._get_api_url_components()
        resource = settings.API_ENDPOINTS.get('supported-currencies')
        path = f'{path}/{self._clean_api_component(resource)}'
        url = URL(scheme=scheme, host=host, path=path)
        currencies = None
        try:
            utils.validate_url(url.url)
            currencies = super(CoindeskAPIClient, self).get(url.url, False)
        except Exception as err:
            msg = err.args[0]
            logger.warning(f'[CoindeskAPICient] Get currencies error. {msg}.')

        if currencies: utils.validate_currencies_settings(currencies)
        return currencies if currencies else utils.get_currencies_settings()

    def get(self, raw: bool = False):
        """
        Make http get request to Coindesk API.

        :param bool raw: enable/disable api response parsing.
        :return *: api http raw response or data.
        """
        try:
            return super(CoindeskAPIClient, self).get(self.url, raw)
        except Exception as err:
            msg = err.args[0]
            logger.error(f'[CoindeskAPICient] API call error. {msg}.')
            raise CoindeskAPIClientError(msg)


class CoindeskAPIHttpResponse(object):
    """
    Enable Coindesk API response data parsing.
    """

    def __init__(self, response: dict = None):
        """
        Initialize Coindesk API http response.

        :param dict response: response data from Coindesk API.
        """
        self.response = response if response is not None else {}
        for key, value in response.items():
            setattr(self, key, value)

    def __str__(self):
        """
        Represent class via params string.

        :return str: class representation.
        """
        classname = self.__class__.__name__
        return f'<{classname} - Coindesk api response>'

    @classmethod
    def parse(cls, response: dict = None, data_type: str = None, currency: str = None):
        """
        Parse http response from Coindesk API.

        :param dict response: response data from Coindesk API.
        :param str data_type: type of data to fetch (currentprice, historical).
        :param str currency: currency to fetch data in.
        """
        if isinstance(response, (dict, str)):
            try:
                if not isinstance(response, dict):
                    response = json.loads(response)
            except (OverflowError, TypeError) as err:
                msg = f'Could not decode response. {err.args[0]}.'
                logger.error(f'[CoindeskAPIHttpResponse] Response error. {msg}')
                raise CoindeskAPIHttpResponseError(msg)
        else:
            msg = 'Response data type must be dict or str.'
            logger.error(f'[CoindeskAPIHttpResponse] Response error. {msg}')
            raise CoindeskAPIHttpResponseError(msg)

        schema = utils.get_schema(data_type, currency)
        cls._validate_response(response, schema)
        return cls(response)

    @staticmethod
    def _validate_response(response: dict, schema: dict):
        """
        Validate http response from Coindesk API.

        :param dict response: json response from API.
        :param dict schema: response schema to validate.
        """
        try:
            return jsonschema.validate(response, schema)
        except (SchemaError, ValidationError) as err:
            msg = err.args[0]
            logger.error(f'[CoindeskAPIHttpResponse] Response error. {msg}.')
            raise CoindeskAPIHttpResponseError(msg)

    def get_response_item(self, item: str):
        """
        Get Coindesk response particular item value.

        :param str item: response item name.
        :return *: corresponding response item value.
        """
        if item not in self.response.keys():
            msg = f'Provided response item {item} does not exist.'
            logger.warning(f'[CoindeskAPIHttpResponse] Invalid response item. {msg}')
        return vars(self).get(item)

    @property
    def response_items(self):
        """
        Get Coindesk response items list.

        :return list: list of response items.
        """
        return list(self.response.keys())

    @property
    def json_response(self):
        """
        Get json response from Coindesk API.

        :return json: json serialized response.
        """
        try:
            return json.dumps(self.response)
        except (OverflowError, TypeError) as err:
            msg = f'Could not encode json response. {err.args[0]}.'
            logger.error(f'[CoindeskAPIHttpResponse] Response error. {msg}')
            raise CoindeskAPIHttpResponseError(msg)



# encoding: utf-8

class BaseError(Exception):
    """
    Handle generic exception.
    """

    def __init__(self, message: str = None, code: int = None):
        """
        Initialize Base error class.

        :param str message: error description message.
        :param int code: error code number.
        """
        if code is not None:
            message = f'Error {code}: {message}.'

        super(BaseError, self).__init__(message)
        self.code = code
        self.message = message


class CoindeskAPIClientError(BaseError):
    """
    Handle exception for Coindesk API client.
    """
    pass


class CoindeskAPIHttpRequestError(BaseError):
    """
    Handle exception for Coindesk API request.
    """
    pass


class CoindeskAPIHttpResponseError(BaseError):
    """
    Handle exception for Coindesk API response.
    """
    pass



# encoding: utf-8

import codecs
import json
import re
from datetime import datetime
from logging import getLogger
from logging.config import fileConfig
from os.path import dirname, join

from . import schemas, settings
from .exceptions import (CoindeskAPIClientError,
                         CoindeskAPIHttpRequestError,
                         CoindeskAPIHttpResponseError)

# Custom logger for client config module
fileConfig(join(dirname(dirname(__file__)), 'logging.cfg'))
logger = getLogger(__name__)


def validate_data_type(data_type: str):
    """
    Validate data type to request.

    :param str data_type: type of data to fetch (currentprice, historical).
    """
    if data_type not in settings.VALID_DATA_TYPES:
        msg = f'Data must be {" or ".join(settings.VALID_DATA_TYPES)}.'
        logger.error(f'[CoinDeskAPIClient] Data error. {msg}')
        raise CoindeskAPIClientError(msg)
    return data_type


def validate_params(data_type: str, params: dict):
    """
    Validate query parameters of the request.

    :param str data_type: type of data to fetch (currentprice, historical).
    :param dict params: optional query parameters.
    :return dict: validated optional query parameters.
    """
    if data_type == settings.API_CURRENTPRICE_DATA_TYPE:
        for param in params.keys():
            if param not in settings.VALID_CURRENTPRICE_PARAMS:
                msg = f'Unvalid param {param} for {data_type} data.'
                logger.error(f'[CoinDeskAPIClient] Param error. {msg}')
                raise CoindeskAPIClientError(msg)
        if settings.CURRENCY_PARAM in params:
            validate_currency(params[settings.CURRENCY_PARAM])
    elif data_type == settings.API_HISTORICAL_DATA_TYPE:
        for param in params.keys():
            if param not in settings.VALID_HISTORICAL_PARAMS:
                msg = f'Unvalid param {param} for {data_type} data.'
                logger.error(f'[CoinDeskAPIClient] Param error. {msg}')
                raise CoindeskAPIClientError(msg)
        if settings.INDEX_PARAM in params:
            validate_index(params=params)
        if settings.CURRENCY_PARAM in params:
            validate_currency(params=params)
        if settings.START_PARAM in params:
            validate_date(params=params, flag=settings.START_PARAM)
        if settings.END_PARAM in params:
            validate_date(params=params, flag=settings.END_PARAM)
        if settings.FOR_PARAM in params:
            validate_for(params=params)
    else:
        msg = f'Unable to validate params for data {data_type}.'
        logger.error(f'[CoinDeskAPIClient] Data error. {msg}')
        raise CoindeskAPIClientError(msg)
    return params


def validate_index(index: str = None, params: dict = None):
    """
    Validate index query parameter.

    :param str index: index optional url query parameter.
    :param dict params: optional query parameters.
    """
    if params is None: params = {}
    index = index or params.get(settings.INDEX_PARAM)
    if index not in settings.VALID_INDEX:
        msg = f'Index must be {" or ".join(settings.VALID_INDEX)}.'
        logger.error(f'[CoinDeskAPIClient] Index error. {msg}')
        raise CoindeskAPIClientError(msg)


def validate_currency(currency: str = None, params: dict = None):
    """
    Validate currency to request data in.

    :param str currency: currency to fetch data in.
    :param dict params: optional query parameters.
    """
    if params is None: params = {}
    currency = currency or params.get(settings.CURRENCY_PARAM)
    currencies = list(map(lambda c: c['currency'], get_currencies_settings()))
    if currency is not None and currency not in currencies:
        msg = f'Unvalid provided currency {currency}.'
        logger.error(f'[CoinDeskAPIClient] Currency error. {msg}')
        raise CoindeskAPIClientError(msg)


def validate_date(date: str = None, params: dict = None, flag: str = None):
    """
    Validate date query parameter.

    :param str date: datetime at which to start/end the chart.
    :param dict params: optional query parameters.
    :param str flag: signalize "start" or "end" date.
    """
    if params is None: params = {}
    date = date or params.get(flag)
    match = re.search(r'^(?P<date>\d{4}-\d{1,2}-\d{1,2})$', date)
    if match:
        try:
            date = datetime.strptime(match.group('date'), '%Y-%m-%d')
            params[flag] = date.strftime('%Y-%m-%d')
        except ValueError as err:
            msg = f'Unable to parse {flag} param. {err.args[0]}.'
            logger.error(f'[CoinDeskAPIClient] Date error. {msg}')
            raise CoindeskAPIClientError(msg)
    else:
        msg = f'{flag.capitalize()} must fullfill the pattern YYYY-MM-DD.'
        logger.error(f'[CoinDeskAPIClient] Date error. {msg}')
        raise CoindeskAPIClientError(msg)


def validate_for(for_param: str = None, params: dict = None):
    """
    Validate for query parameter.

    :param str for: datetime signaling previous day data.
    :param dict params: optional query parameters.
    """
    if params is None: params = {}
    for_param = for_param or params.get(settings.FOR_PARAM)
    if for_param not in settings.VALID_FOR:
        msg = f'For must be {settings.VALID_FOR[0]}.'
        logger.error(f'[CoinDeskAPIClient] For error. {msg}')
        raise CoindeskAPIClientError(msg)


def validate_retries(retries: int):
    """
    Validate request retries attempt parameter.

    :param int retries: number of request attempts before failing.
    :return int: retries number.
    """
    if type(retries) is not int or retries < 0:
        msg = 'Retries must be zero or positive integer number.'
        logger.error(f'[CoindeskAPIHttpRequest] Retries error. {msg}')
        raise CoindeskAPIHttpRequestError(msg)
    max_retries = min(retries, settings.REQUEST_MAX_RETRIES)
    if max_retries < retries:
        logger.warning(f'[CoinDeskAPIClient] Request max retries. {max_retries}.')
    return max_retries


def validate_redirects(redirects: bool):
    """
    Validate http request redirects.

    :param bool redirects: enable/disable http verbs redirection.
    :return bool: redirects status.
    """
    if not isinstance(redirects, bool):
        msg = 'Redirects must be boolean.'
        logger.error(f'[CoindeskAPIHttpRequest] Redirects error. {msg}')
        raise CoindeskAPIHttpRequestError(msg)
    return redirects


def validate_timeout(timeout: int):
    """
    Validate request timeout.

    :param int timeout: number seconds of request timeout.
    :return int: timeout seconds number.
    """
    if type(timeout) is not int or timeout < 0:
        msg = 'Timeout must be zero or positive integer number.'
        logger.error(f'[CoindeskAPIHttpRequest] Timeout error. {msg}')
        raise CoindeskAPIHttpRequestError(msg)
    max_timeout = min(timeout, settings.REQUEST_MAX_TIMEOUT)
    if max_timeout < timeout:
        logger.warning(f'[CoinDeskAPIClient] Request max retries. {max_timeout}.')
    return max_timeout


def validate_backoff(backoff: bool):
    """
    Validate http request backoff.

    :param bool backoff: enable/disable http request backoff.
    :return bool: backoff status.
    """
    if not isinstance(backoff, bool):
        msg = 'Backoff must be boolean.'
        logger.error(f'[CoindeskAPIHttpRequest] Backoff error. {msg}')
        raise CoindeskAPIHttpRequestError(msg)
    return backoff


def validate_url(url: str):
    """
    Validate Coindesk constructed url.

    :param str url: Coindesk api endpoint.
    """
    regex = re.compile(
        r'^(?:http)s?://'
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?))'
        r'(?:[/?#][^\s]*)?$', re.IGNORECASE)
    match = regex.search(url)
    if not match:
        msg = f'Invalid url {url}.'
        logger.error(f'[CoinDeskAPIClient] Url error. {msg}')
        raise CoindeskAPIClientError(msg)
    return url


def validate_currencies_settings(currencies: list):
    """
    Validate supported currencies settings against feched ones.

    :param list currencies: Coindesk API supported currencies.
    """
    codes = list(map(lambda c: c['currency'], get_currencies_settings()))
    supported_codes = list(map(lambda c: c['currency'], currencies))
    if not set(codes) == set(supported_codes):
        msg = 'Valid currencies settings out of date.'
        logger.warning(f'[CoindeskAPIHttpRequest] Currencies warn. {msg}')
        update_currencies_settings(currencies)


def get_currencies_settings():
    """
    Get supported currencies settings list.

    :return list: valid currencies list from settings file.
    """
    currencies_json = join(dirname(__file__), 'currencies.json')
    try:
        with open(currencies_json) as currencies_file:
            currencies = json.load(currencies_file)
    except (OSError, IOError) as err:
        msg = f'Unable to read currencies file. {err.args[0]}.'
        logger.error(f'[CoindeskAPIClient] File error. {msg}')
        raise CoindeskAPIClientError(msg)
    return currencies.get('SUPPORTED_CURRENCIES')


def update_currencies_settings(currencies: list):
    """
    Get supported currencies settings list.

    :param list currencies: Coindesk API supported currencies.
    """
    currencies_json = join(dirname(__file__), 'currencies.json')
    supported_currencies = {'SUPPORTED_CURRENCIES': currencies}
    try:
        with open(currencies_json, 'w') as outfile:
            json.dump(supported_currencies, outfile, indent=2)
    except (OSError, IOError) as err:
        msg = f'Unable to write currencies file. {err.args[0]}.'
        logger.error(f'[CoindeskAPIClient] File error. {msg}')
        raise CoindeskAPIClientError(msg)
    msg = 'Currencies file successfully updated.'
    logger.info(f'[CoindeskAPIClient] File updated. {msg}')


def get_schema(data_type: str, currency: str = None):
    """
    Get schema for corresponding data and currency.

    :param str data_type: type of data to fetch (currentprice, historical).
    :param str currency: code for allowed currency.
    :return dict: CoinDesk API response schema.
    """
    schema = {}
    if data_type == settings.API_CURRENTPRICE_DATA_TYPE:
        if currency is not None:
            schema = get_schema_for_currency(currency)
        else:
            schema = schemas.CURRENTPRICE_SCHEMA
    elif data_type == settings.API_HISTORICAL_DATA_TYPE:
        schema = schemas.HISTORICAL_SCHEMA

    if not schema:
        msg = f'No schema for data type {data_type} and currency {currency}.'
        logger.error(f'[CoinDeskAPIHttpResponse] Schema error. {msg}')
        raise CoindeskAPIHttpResponseError(msg)
    return schema


def get_schema_for_currency(currency: str):
    """
    Set currentprice schema for specific currency.

    :param str currency: code for allowed currency.
    :return dict: Updated schema for currency.
    """
    schema = schemas.CURRENTPRICE_CODE_SCHEMA
    schema["properties"]["bpi"].update({
        currency: {
            "code": {"type": "string"},
            "symbol": {"type": "string"},
            "description": {"type": "string"},
            "rate_float": {"type": "number"}
        }
    })
    return schema


def find_version(*path_parts):
    """
    Get current coindesk api client version.

    :param tuple path_parts: path parts for version module.
    :return str: api client version.
    """
    version = read(*path_parts)
    match = re.search(r'^__version__ = ["\'](?P<version>[^"\']*)["\']', version, re.M)
    if not match:
        msg = 'Unable to find module version.'
        logger.error(f'[find_version] Setup error. {msg}')
        raise RuntimeError(msg)
    return str(match.group('version'))


def read(*path_parts):
    """
    Get version module.

    :param tuple path_parts: path parts for version module.
    :return obj: module with version.
    """
    base_dir = dirname(__file__)
    file_path = join(base_dir, *path_parts)
    return codecs.open(file_path, encoding='utf-8').read()



# encoding: utf-8

# Coindesk API response schemas
CURRENTPRICE_SCHEMA = {
    "type": "object",
    "properties": {
        "time": {
            "updated": {"type": "string"},
            "updatedISO": {"type": "string"},
            "updateduk": {"type": "string"}
        },
        "chartName": {"type": "string"},
        "disclaimer": {"type": "string"},
        "bpi": {
            "USD": {
                "code": {"type": "string"},
                "symbol": {"type": "string"},
                "rate": {"type": "string"},
                "description": {"type": "string"},
                "rate_float": {"type": "number"}
            },
            "GBP": {
                "code": {"type": "string"},
                "symbol": {"type": "string"},
                "rate": {"type": "string"},
                "description": {"type": "string"},
                "rate_float": {"type": "number"}
            },
            "EUR": {
                "code": {"type": "string"},
                "symbol": {"type": "string"},
                "rate": {"type": "string"},
                "description": {"type": "string"},
                "rate_float": {"type": "number"}
            }
        }
    },
    "required": ["chartName", "time", "disclaimer", "bpi"]
}

CURRENTPRICE_CODE_SCHEMA = {
    "type": "object",
    "properties": {
        "time": {
            "updated": {"type": "string"},
            "updatedISO": {"type": "string"},
            "updateduk": {"type": "string"}
        },
        "disclaimer": {"type": "string"},
        "bpi": {
            "USD": {
                "code": {"type": "string"},
                "rate": {"type": "string"},
                "description": {"type": "string"},
                "rate_float": {"type": "number"}
            }
        }
    },
    "required": ["time", "disclaimer", "bpi"]
}

HISTORICAL_SCHEMA = {
    "type": "object",
    "properties": {
        "time": {
            "updated": {"type": "string"},
            "updatedISO": {"type": "string"}
        },
        "disclaimer": {"type": "string"},
        "bpi": {"type": "object"}
    },
    "required": ["time", "disclaimer", "bpi"]
}



# encoding: utf-8

import asyncio
from functools import wraps
from logging import getLogger
from logging.config import fileConfig
from os.path import dirname, join

# Custom logger for decorators module
fileConfig(join(dirname(dirname(__file__)), 'logging.cfg'))
logger = getLogger(__name__)


def async_event_loop(method):
    @wraps(method)
    def result(self, *args, **kwargs):
        caller = self.__class__.__name__
        logger.info(f'[{caller}] Event loop initialized.')
        response = asyncio.run(method(self, *args, **kwargs))
        logger.info(f'[{caller}] Event loop finished.')
        return response
    return result



#!/usr/bin/env python
# encoding: utf-8

import codecs
import re
from os.path import dirname, join
from setuptools import setup


def find_version(*path_parts):
    """
    Get current coindesk api client version.

    :param tuple path_parts: path parts for version module.
    :return str: api client version.
    """
    version = read(*path_parts)
    match = re.search(r'^__version__ = ["\'](?P<version>[^"\']*)["\']', version, re.M)
    if not match:
        msg = 'Unable to find module version.'
        raise RuntimeError(msg)
    return str(match.group('version'))


def read(*path_parts):
    """
    Get version module.

    :param tuple path_parts: path parts for version module.
    :return obj: module with version.
    """
    base_dir = dirname(__file__)
    file_path = join(base_dir, *path_parts)
    return codecs.open(file_path, encoding='utf-8').read()


setup(
    name='coindesk',
    version=find_version('coindesk', '__init__.py'),
    description='Bitcoin current and historical price',
    long_description=read('README.md'),
    long_description_content_type="text/markdown",
    author='Sergio de Diego',
    author_email='sergiodediego@outlook.com',
    url='https://github.com/sdediego/coindesk-api-client',
    download_url='https://github.com/sdediego/coindesk-api-client/archive/v1.1.0-alpha.tar.gz',
    packages=['coindesk'],
    install_requires=[
        "aiohttp>=3.6.1",
        "jsonschema>=3.0.2",
        "requests>=2.22.0",
        "furl>=2.1.0",
    ],
    license='MIT',
    keywords='api asynchronous Bitcoin blockchain client Coindesk Python',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Financial and Insurance Industry',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Topic :: Office/Business :: Financial',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)
